"""Extract meta-data from a font-file to describe the font"""
from fontTools import ttLib
import sys
try:
	from OpenGLContext.debug.logs import text_log
except ImportError:
	text_log = None

def openFont( filename ):
	"""Get a new font object"""
	return ttLib.TTFont(filename)

FONT_SPECIFIER_NAME_ID = 4
FONT_SPECIFIER_FAMILY_ID = 1
def shortName( font ):
	"""Get the short name from the font's names table"""
	name = ""
	family = ""
	for record in font['name'].names:
		if record.nameID == FONT_SPECIFIER_NAME_ID and not name:
			if '\000' in record.string:
				name = unicode(record.string, 'utf-16-be').encode('utf-8')
			else:
				name = record.string
		elif record.nameID == FONT_SPECIFIER_FAMILY_ID and not family:
			if '\000' in record.string:
				family = unicode(record.string, 'utf-16-be').encode('utf-8')
			else:
				family = record.string
		if name and family:
			break
	return name, family


FAMILY_NAMES = {
	0: ("ANY",{}),
	1: ("SERIF-OLD", {
		0: "ANY",
		1: "ROUNDED-LEGIBILITY",
		2: "GARALDE",
		3: "VENETIAN",
		4: "VENETIAN-MODIFIED",
		5: "DUTCH-MODERN",
		6: "DUTCH-TRADITIONAL",
		7: "CONTEMPORARY",
		8: "CALLIGRAPHIC",
		15: "MISCELLANEOUS",
	}),
	2: ("SERIF-TRANSITIONAL", {
		0: "ANY",
		1: "DIRECT-LINE",
		2: "SCRIPT",
		15: "MISCELLANEOUS",
	}),
	3: ("SERIF", {
		0: "ANY",
		1: "ITALIAN",
		2: "SCRIPT",
		15: "MISCELLANEOUS",
	}),
	4: ("SERIF-CLARENDON",{
		0: "ANY",
		1: "CLARENDON",
		2: "MODERN",
		3: "TRADITIONAL",
		4: "NEWSPAPER",
		5: "STUB-SERIF",
		6: "MONOTYPE",
		7: "TYPEWRITER",
		15: "MISCELLANEOUS",
	}),
	5: ("SERIF-SLAB",{
		0: 'ANY',
		1: 'MONOTONE',
		2: 'HUMANIST',
		3: 'GEOMETRIC',
		4: 'SWISS',
		5: 'TYPEWRITER',
		15: 'MISCELLANEOUS',
	}),
	7: ("SERIF-FREEFORM",{
		0: 'ANY',
		1: 'MODERN',
		15: 'MISCELLANEOUS',
	}),
	8: ("SANS",{
		0: 'ANY',
		1: 'GOTHIC-NEO-GROTESQUE-IBM',
		2: 'HUMANIST',
		3: 'ROUND-GEOMETRIC-LOW-X',
		4: 'ROUND-GEOMETRIC-HIGH-X',
		5: 'GOTHIC-NEO-GROTESQUE',
		6: 'GOTHIC-NEO-GROTESQUE-MODIFIED',
		9: 'GOTHIC-TYPEWRITER',
		10: 'MATRIX',
		15: 'MISCELLANEOUS',
	}),
	9: ("ORNAMENTAL",{
		0: 'ANY',
		1: 'ENGRAVER',
		2: 'BLACK-LETTER',
		3: 'DECORATIVE',
		4: 'THREE-DIMENSIONAL',
		15: 'MISCELLANEOUS',
	}),
	10:("SCRIPT",{
		0: 'ANY',
		1: 'UNCIAL',
		2: 'BRUSH-JOINED',
		3: 'FORMAL-JOINED',
		4: 'MONOTONE-JOINED',
		5: 'CALLIGRAPHIC',
		6: 'BRUSH-UNJOINED',
		7: 'FORMAL-UNJOINED',
		8: 'MONOTONE-UNJOINED',
		15: 'MISCELLANEOUS',
	}),
	12:("SYMBOL",{
		0: 'ANY',
		3: 'MIXED-SERIF',
		6: 'OLDSTYLE-SERIF',
		7: 'NEO-GROTESQUE-SANS',
		15: 'MISCELLANEOUS',
	}),
}

WEIGHT_NAMES = {
	'thin':100,
	'extralight':200,
	'ultralight':200,
	'light':300,
	'normal':400,
	'regular':400,
	'plain':400,
	'medium':500,
	'semibold':600,
	'demibold':600,
	'bold':700,
	'extrabold':800,
	'ultrabold':800,
	'black':900,
	'heavy':900,
}
WEIGHT_NUMBERS = {}
for key,value in WEIGHT_NAMES.items():
	WEIGHT_NUMBERS.setdefault(value,[]).append(key)

def weightNumber( name ):
	"""Convert a string-name to a weight number compatible with this module"""
	if isinstance( name, (str,unicode)):
		name = name.lower()
		name = name.replace( '-','').replace(' ','')
		if name and name[-1] == '+':
			name = name[:-1]
			adjust = 50
		elif name and name[-1] == '-':
			name = name[:-1]
			adjust = -50
		else:
			adjust = 0
		return WEIGHT_NAMES[ name ]+ adjust
	else:
		return int(name) or 400 # for cases where number isn't properly specified

def weightName( number ):
	"""Convert integer number to a human-readable weight-name"""
	number = int(number) or 400
	if WEIGHT_NUMBERS.has_key( number ):
		return WEIGHT_NUMBERS[number]
	name = 'thin-'
	for x in range(100,1000, 100):
		if number >= x:
			name = WEIGHT_NUMBERS[x]+'+'
	return name

def family( font ):
	"""Get the family (and sub-family) for a font"""
	HIBYTE = 65280
	LOBYTE = 255
	familyID = (font['OS/2'].sFamilyClass&HIBYTE)>>8
	subFamilyID = font['OS/2'].sFamilyClass&LOBYTE
	return familyNames( familyID, subFamilyID )
def familyNames( familyID, subFamilyID=0 ):
	"""Convert family integers to human-readable names"""
	familyName, subFamilies = FAMILY_NAMES.get( familyID, ('RESERVED',None))
	if familyName == 'RESERVED':
		if text_log:
			text_log.warn( 'Font has invalid (reserved) familyID: %s', familyID )
	if subFamilies:
		subFamily = subFamilies.get(subFamilyID, 'RESERVED')
	else:
		subFamily = 'ANY'
	return (
		familyName,
		subFamily
	)

def modifiers( font ):
	"""Get weight and italic modifiers for a font

	weight is taken from the OS/2 usWeightClass field
	italic is taken from either OS/2 fsSelection or
		head macStyle, if either indicates italics we
		report italics
	"""
	return (
		# weight as an integer
		font['OS/2'].usWeightClass,
		( # italic
			font['OS/2'].fsSelection &1 or
			font['head'].macStyle&2
		),
	)

def guessEncoding( font, given=None ):
	"""Attempt to guess/retrieve an encoding from the font itself

	Basically this will try to get the given encoding
	(unless it is None).

	If given is a single integer or a single-item tuple,
	we will attempt scan looking for any table matching
	given as the platform ID and returning the first sub
	table.

	If given is a two-value tuple, we will require
	explicit matching, and raise errors if the encoding
	cannot be retrieved.

	if given is None, we will return the first encoding
	in the font.

	XXX This needs some work, particularly for non-win32
		platforms, where there is no preference embodied
		for the native encoding.
	"""
	if isinstance( given, tuple) and given:
		if len(given) == 2:
			if __debug__:
				if text_log:
					text_log.info(
						"""Checking for explicitly required encoding %r""",
						given
					)
			if not font['cmap'].getcmap( *given ):
				raise ValueError(
					"""The specified font encoding %r does not appear to be available within the font %r. Available encodings: %s"""% (
						given, shortName(font),
						[
							(table.platformID,table.platEncID)
							for table in font['cmap'].tables
						],
					)
				)
			return given
		elif len(given) > 2:
			raise TypeError("""Encoding must be None, a two-tuple, or an integer, got %r"""%(given,))
		else:
			# treat as a single integer, regardless of number of integer's
			given = given[0]
	if isinstance( given, (int, long)):
		for table in font['cmap'].tables:
			if table.platformID == given:
				return (table.platformID, table.platEncID)
		raise ValueError(
			"""Could not find encoding with specified platformID==%s within the font %r. Available encodings: %s"""% (
				given, shortName(font),
				[
					(table.platformID,table.platEncID)
					for table in font['cmap'].tables
				],
			)
		)
	if sys.platform == 'win32':
		prefered = (3,1)
		# should have prefered values for Linux and Mac as well...
		for table in font['cmap'].tables:
			if (table.platformID, table.platEncID) == prefered:
				return prefered
	# just retrieve the first table's values
	for table in font['cmap'].tables:
		return (table.platformID, table.platEncID)
	raise ValueError(
		"""There are no encoding tables within the font %r, likely a corrupt font-file"""% (
			shortName(font),
		)
	)



"""Representation of a single glyph including contour extraction"""
from ttfquery import glyphquery
from Numeric import *

class Glyph( object):
	"""Object encapsulating metadata regarding a particular glyph"""
	def __init__(self, glyphName ):
		"""Initialize the glyph object

		glyphName -- font's glyphName for this glyph, see
			glyphquery.glyphName() for extracting the
			appropriate name from character and encoding.
		"""
		self.glyphName = glyphName
	def compile( self, font, steps = 3 ):
		"""Compile the glyph to a set of poly-line outlines"""
		self.contours = self.calculateContours( font )
		self.outlines = [
			decomposeOutline(contour,steps)
			for contour in self.contours
		]
		self.width = glyphquery.width( font, self.glyphName )
		self.height = glyphquery.lineHeight( font )
	
	def calculateContours( self, font ):
		"""Given a character, determine contours to draw

		returns a list of contours, with each contour
		being a list of ((x,y),flag) elements.  There may
		be only a single contour.
		"""
		glyf = font['glyf']
		charglyf = glyf[self.glyphName]
		return self._calculateContours( charglyf, glyf, font )
	def _calculateContours( self, charglyf, glyf, font ):
		"""Create expanded contour data-set from TTF charglyph entry

		This is called recursively for composite glyphs, so
		I suppose a badly-formed TTF file could cause an infinite
		loop.

		charglyph -- glyf table's entry for the target character
		glyf -- the glyf table (used for recursive calls)
		"""
		charglyf.expand( font ) # XXX is this extraneous?
		contours = []
		if charglyf.numberOfContours == 0:
			# does not display at all, for instance, space
			pass
		elif charglyf.isComposite():
			# composed out of other items...
			for component in charglyf.components:
				subContours = self._calculateContours(
					glyf[component.glyphName],
					glyf,
					font
				)
				dx,dy = component.getComponentInfo()[1][-2:]
				# XXX we're ignoring the scaling/shearing/etceteras transformation
				# matrix which is given by component.getComponentInfo()[1][:4]
				subContours = [
					[((x+dx,y+dy),f) for ((x,y),f) in subContour]
					for subContour in subContours
				]
				contours.extend(
					subContours
				)
		else:
			flags = charglyf.flags
			coordinates = charglyf.coordinates
			# We need to convert from the "end point" representation
			# to a distinct contour representation, requires slicing
			# each contour out of the list and then closing it
			last = 0
			for e in charglyf.endPtsOfContours:
				set = map(
					None,
					list(coordinates[last:e+1])+list(coordinates[last:last+1]),
					list(flags[last:e+1])+list(flags[last:last+1])
				)
				contours.append( set )
				last = e+1
			if coordinates[last:]:
				contours.append( map(
					None,
					list(coordinates[last:])+list(coordinates[last:last+1]),
					list(flags[last:])+list(flags[last:])
				) )
		return contours
		
def decomposeOutline( contour, steps=3 ):
	"""Decompose a single TrueType contour to a line-loop

	In essence, this is the "interpretation" of the font
	as geometric primitives.  I only support line and
	quadratic (conic) segments, which should support most
	TrueType fonts as far as I know.

	The process consists of first scanning for any multi-
	of-curve control-point runs.  For each pair of such
	control points, we insert a new on-curve control point.

	Once we have the "expanded" control point array we
	scan through looking for each segment which includes
	an off-curve control point.  These should only be
	bracketed by on-curve control points.  For each
	found segment, we call our integrateQuadratic method
	to produce a set of points interpolating between the
	end points as affected by the middle control point.

	All other control points merely generate a single
	line-segment between the endpoints.
	"""
	# contours must be closed, but they can start with
	# (and even end with) items not on the contour...
	# so if we do, we need to create a new point to serve
	# as the midway...
	if len(contour)<3:
		return ()
	set = contour[:]
	def on( ((Ax,Ay),Af) ):
		"""Is this record on the contour?"""
		return Af==1

	def merge( ((Ax,Ay),Af), ((Bx,By),Bf)):
		"""Merge two off-point records into an on-point record"""
		return (((Ax+Bx)/2.0),((Ay+By))/2.0),1
	# create an expanded set so that all adjacent
	# off-curve items have an on-curve item added
	# in between them
	last = contour[-1]
	expanded = []
	for item in set:
		if (not on(item)) and (not on(last)):
			expanded.append( merge(last, item))
		expanded.append( item )
		last = item
	result = []
	last = expanded[-1]
	while expanded:
		assert on(expanded[0]), "Expanded outline doesn't have proper format! Should always have either [on, off, on] or [on, on] as the first items in the outline"
		if len(expanded)>1:
			if on(expanded[1]):
				# line segment from 0 to 1
				result.append( expanded[0][0] )
				#result.append( expanded[1][0] )
				del expanded[:1]
			else:
				assert on(expanded[2]), "Expanded outline doesn't have proper format!"
				points = integrateQuadratic( expanded[:3], steps = steps )
				result.extend( points )
				del expanded[:2]
		else:
			assert on(expanded[0]), """Expanded outline finishes off-curve"""
			result.append( expanded[0][0] )
			del expanded[:1]
	result.append( result[-1] )
	return result
def integrateQuadratic( points, steps=3 ):
	"""Get points on curve for quadratic w/ end points A and C

	Basis Equations are taken from here:
		http://www.truetype.demon.co.uk/ttoutln.htm

	This is a very crude approach to the integration,
	everything is coded directly in Python, with no
	attempts to speed up the process.

	XXX Should eventually provide adaptive steps so that
		the angle between the elements can determine how
		many steps are used.
	"""
	step = 1.0/steps
	((Ax,Ay),_),((Bx,By),_),((Cx,Cy),_) = points
	result = [(Ax,Ay)]
	### XXX This is dangerous, in certain cases floating point error
	## can wind up creating a new point at 1.0-(some tiny number) if step
	## is slightly less than precisely 1.0/steps
	for t in arange( step, 1.0, step ):
		invT = 1.0-t
		px = (invT*invT * Ax) + (2*t*invT*Bx) + (t*t*Cx)
		py = (invT*invT * Ay) + (2*t*invT*By) + (t*t*Cy)
		result.append( (px,py) )
	# the end-point will be added by the next segment...
	#result.append( (Cx,Cy) )
	return result



#!/usr/bin/env python
"""Query for meta-data for a given font"""
from ttfquery import describe

def printMetaData( metadata ):
	print '    Specific Name:', metadata[2]
	print '    File:', metadata[0]
	print '    Modifiers:', metadata[1]
	print '    Family Name:', ", ".join(metadata[4])
	

if __name__ == "__main__":
	usage ="""metadata_query name

	Will create a registry file font.cache if it doesn't
	already exist, otherwise will just use the existing
	cache.  See ttffiles.py for updating the cache.
	"""
	from ttfquery._scriptregistry import registry
	import sys, os
	if sys.argv[1:]:
		name = " ".join( sys.argv[1:] )
	else:
		sys.stderr.write( usage )
		sys.exit(1)
	fontNames = registry.matchName( name )
	fontNames.sort()
	for general in fontNames:
		print 'Font:', general
		specifics = registry.fontMembers( general )
		specifics.sort()
		for specific in specifics:
			printMetaData( registry.metadata( registry.fontFile(specific) ))
	
	




"""Registry of available TrueType font files

XXX Currently two copies of exactly the same font
	will likely confuse the registry because the
	specificFonts set will only have one of the
	metrics sets.  Nothing breaks at the moment
	because of this, but it's not ideal.
"""
from ttfquery import describe, findsystem
import cPickle, time, traceback, os, sys


FILENAME, MODIFIERS, SPECIFICNAME, FONTNAME, FAMILY = range(5)

class Registry(object):
	"""Object providing centralized registration of TTF files

	Attributes:
		families -- mapping from TrueType font families
			to sub-families and then to general fonts
			(as a set of font-names).
		fonts -- mapping from general fonts to modifiers to
			specific font instances
		specificFonts -- mapping from specific font names
			to the entire "metrics" set for the particular
			font.
		files -- mapping from (absolute) filenames to
			specific font names
		shortFiles -- mapping from font filename basenames
			to font-file-lists
		DIRTY -- flag indicating whether the registry has
			had a new font registered (i.e. whether it should
			be saved out to disk).
	"""
	DIRTY = 0
	filename = ""
	def __init__(self):
		"""Initialize the Registry"""
		self.families = {}
		self.fonts = {}
		self.specificFonts = {}
		self.files = {}
		self.shortFiles = {}
	def clear( self ):
		"""Clear out the all tables and mark unchanged"""
		self.families.clear()
		self.fonts.clear()
		self.specificFonts.clear()
		self.files.clear()
		self.shortFiles.clear()
		self.dirty(0)
	def dirty(self, dirty = 1):
		"""Mark the registry as changed/unchanged"""
		self.DIRTY = dirty

	def metadata(
		self,
		filename,
		force = 0
	):
		"""Retrieve metadata from font file

		filename -- fully specified path to the font file
		force -- if false, and the metadata is already
			available for this file, do not access the
			font file to retrieve, just return the existing
			metadata.

		return value:
			tuple of:
				filename -- fully specified absolute path
				modifiers -- (weightInteger, italicsFlag)
				specificName -- specific name of the particular
					font stored in the given file, the name of
					the "modified" font
				fontName -- name of the general font which
					the modifiers are specialising
				specifier -- family specifier, two-tuple of
					high-level and sub-level font classifications
					based on font characteristics as encoded
					in the font file.
		"""
		filename = os.path.abspath( filename )
		if self.files.has_key( filename ) and not force:
			return self.specificFonts.get( self.files[filename] )
		font = describe.openFont(filename)
		try:
			modifiers = describe.modifiers( font )
		except (KeyError,AttributeError), err:
			modifiers = (None,None)
		specificName, fontName = describe.shortName( font )
		specifier = describe.family(font)
		return (
			filename,
			modifiers,
			specificName,
			fontName,
			specifier,
		)
	def register(
		self,
		filename,
		modifiers = None,
		specificName = None,
		fontName = None,
		familySpecifier = None,
		force = 0,
	):
		"""Do the actual registration of a filename & metadata

		See metadata function for description of the various
		arguments.  If modifiers == None then the metadata function
		will be used to scan for the metadata.

		force -- if true, force re-reading font-file even if we already
			have the meta-data for the file loaded.
		"""
		filename = os.path.abspath( filename )
		if self.files.has_key( filename ) and not force:
			return self.specificFonts.get( self.files[filename] )
		self.dirty(1)
		if modifiers == None:
			(filename, modifiers, specificName, fontName, familySpecifier) = self.metadata(filename, force = force)
		description = (filename, modifiers, specificName, fontName, familySpecifier)
		try:
			self.files[filename] = specificName
			major,minor = familySpecifier
			self.families.setdefault(major,{}).setdefault(minor,{})[fontName] = 1
			self.fonts.setdefault(fontName, {}).setdefault(modifiers,[]).append(specificName)
			self.specificFonts[ specificName ] = description
			self.shortFiles.setdefault(os.path.basename(filename), []).append( filename )
		except Exception:
			if self.files.has_key(filename):
				del self.files[filename]
			raise
		return description

	def familyMembers( self, major, minor=None ):
		"""Get all (general) fonts for a given family"""
		major = major.upper()
		if not minor:
			result = []
			for key,set in self.families.get(major,{}).items():
				result.extend( set.keys())
			return result
		minor = minor.upper()
		return self.families.get( major, {}).get(minor,{}).keys()
	def fontMembers( self, fontName, weight=None, italics=None ):
		"""Get specific font names for given generic font name

		weight -- if specified, only members with the given weight
		italics -- if specified, only members where the flag matches

		returns list of specific font names
		"""
		table = self.fonts.get( fontName, {})
		items = table.items()
		items.sort()
		if weight is not None:
			weight = describe.weightNumber( weight )
			items = [item for item in items if item[0][0]==weight]
		if italics is not None:
			items = [item for item in items if item[0][1]==italics]
		result = []
		for item in items:
			result.extend( item[1])
		return result
	def fontForms( self, fontName ):
		"""Retrieve the set of font-forms (weight,italics) available in a font"""
		return self.fonts.get( fontName, {}).keys()

	def fontFile( self, specificName ):
		"""Return the absolute path-name for a given specific font"""
		description = self.specificFonts.get( specificName )
		if description:
			return description[0]
		else:
			raise KeyError( """Couldn't find font with specificName %r, can't retrieve filename for it"""%( specificName,))

	def matchName( self, name, single=0 ):
		"""Try to find a general font based on a name"""
		result = {}
		if self.fonts.has_key( name ):
			v = name
			if single:
				return v
			else:
				result[v] = 1
		if self.specificFonts.has_key( name ):
			v = self.specificFonts[name][FONTNAME]
			if single:
				return v
			else:
				result[v] = 1
		if self.families.has_key( name.upper() ):
			for general in self.familyMembers( name ):
				if single:
					return general
				result[general] = 1
		testname = name.lower()
		for specific in self.specificFonts.keys():
			if specific.lower().find( testname ) > -1:
				if single:
					return specific
				result[ self.specificFonts[specific][FONTNAME]]=1
		for majorFamily in self.families.keys():
			if majorFamily.lower().find( testname ) > -1:
				for item in self.familyMembers( majorFamily ):
					if single:
						return item
					result[item] = 1
			else:
				# if previous was true, we already included everything
				# that could be included here...
				for minorFamily in self.families[majorFamily].keys():
					if minorFamily.lower().find( testname ) > -1:
						for item in self.familyMembers( majorFamily, minorFamily ):
							if single:
								return item
							result[item] = 1
		if not result:
			raise KeyError( """Couldn't find a font with name %r"""%(name,))
		return result.keys()

	def save( self, file= None, force=0 ):
		"""Attempt to save the font metadata to a pickled file

		file -- a file open in binary write mode or a filename
		force -- if not true and DIRTY false, then don't actually
			save anything

		returns number of records saved
		"""
		if not force and not self.DIRTY:
			return 0
		file = file or self.filename
		if not file:
			raise TypeError( """Attempted to save %r to default file, no default file specified"""% (self,))
		if not hasattr( file, 'write'):
			file = open( file, 'wb' )
		cPickle.dump( self.specificFonts.values(), file, 1 )
		return len(self.specificFonts)
	def load( self, file, clearFirst=1 ):
		"""Attempt to load the font metadata from a pickled file

		file -- a file open in binary read mode or a filename
		clearFirst -- if true, clear tables first, and reset DIRTY
			to 0 after finished
		"""
		if clearFirst:
			self.clear()
		if not hasattr( file, 'read'):
			self.filename = file
			file = open( file, 'rb' )
		table = cPickle.load( file )
		for filename, modifiers, specificName, fontName, familySpecifier in table:
			## Minimal sanity check...
			if os.path.isfile( filename ):
				self.register(filename, modifiers, specificName, fontName, familySpecifier)
		if clearFirst:
			self.dirty(0)
		return len(table)

	def scan( self, paths=None, printErrors=0, force = 0 ):
		"""Scan the given paths registering each found font"""
		new, failed = [],[]
		for filename in findsystem.findFonts(paths):
			try:
				self.register( filename, force = force )
			except Exception, err:
				if printErrors:
					traceback.print_exc()
				failed.append( filename )
			else:
				new.append( filename )
		return new, failed

def load( *arguments, **named ):
	"""Construct registry from saved file

	Assembly creates a Registry object and calls the
	load method with the file as argument.
	"""
	registry = Registry()
	registry.load( *arguments, **named )
	return registry


if __name__ == "__main__":
	usage ="""ttffiles [registryFile [directories]]

	Update registryFile (default "font.cache") by scanning
	the given directories, the system font directories by
	default.
	"""
	import sys
	if sys.argv[1:2]:
		testFilename = sys.argv[1]
		if sys.argv[2:]:
			directories = sys.argv[2:]
		else:
			directories = None
	else:
		testFilename = "font.cache"
		directories = None
	if os.path.isfile( testFilename ):
		registry = load( testFilename )
	else:
		registry = Registry()
	new,failed = registry.scan( directories, printErrors = 1, force = 0)
	for f in failed:
		print 'FAIL', f
	print '%s fonts available'%( len(new), )
	registry.save(testFilename)
	



"""Find system fonts (only works on Linux and Win32 at the moment)"""
import sys, os, glob, re

def win32FontDirectory( ):
	"""Get User-specific font directory on Win32"""
	try:
		import _winreg
	except ImportError:
		return os.path.join(os.environ['WINDIR'], 'Fonts')
	else:
		k = _winreg.OpenKey(
			_winreg.HKEY_CURRENT_USER,
			r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
		)
		try:
			# should check that k is valid? How?
			return _winreg.QueryValueEx( k, "Fonts" )[0]
		finally:
			_winreg.CloseKey( k )

def win32InstalledFonts( fontDirectory = None ):
	"""Get list of explicitly *installed* font names"""
	import _winreg
	if fontDirectory is None:
		fontDirectory = win32FontDirectory()
	if os.name == 'nt':
		keyName = r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts"
	else:
		keyName = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Fonts"
	items = {}
	k = _winreg.OpenKey(
		_winreg.HKEY_LOCAL_MACHINE,
		keyName
	)
	try:
		# should check that k is valid? How?
		for index in range( _winreg.QueryInfoKey(k)[1]):
			key,value,_ = _winreg.EnumValue( k, index )
			if not os.path.dirname( value ):
				value = os.path.join( fontDirectory, value )
			value = os.path.abspath( value ).lower()
			if value[-4:] == '.ttf':
				items[ value ] = 1
		return items.keys()
	finally:
		_winreg.CloseKey( k )
	

def linuxFontDirectories( ):
	"""Get system font directories on Linux/Unix
	
	Uses /usr/sbin/chkfontpath to get the list
	of system-font directories, note that many
	of these will *not* be truetype font directories.

	If /usr/sbin/chkfontpath isn't available, uses
	returns a set of common Linux/Unix paths
	"""
	executable = '/usr/sbin/chkfontpath'
	if os.path.isfile( executable ):
		data = os.popen( executable ).readlines()
		match = re.compile( '\d+: (.+)')
		set = []
		for line in data:
			result = match.match( line )
			if result:
				set.append(result.group(1))
		return set
	else:
		directories = [
			# what seems to be the standard installation point
			"/usr/X11R6/lib/X11/fonts/TTF/",
			# common application, not really useful
			"/usr/lib/openoffice/share/fonts/truetype/",
			# documented as a good place to install new fonts...
			"/usr/share/fonts",
			# okay, now the OS X variants...
			"~/Library/Fonts/",
			"/Library/Fonts/",
			"/Network/Library/Fonts/",
			"/System/Library/Fonts/",
			"System Folder:Fonts:",
		]
		
		set = []
		def add( arg, directory, files):
			set.append( directory )
		for directory in directories:
			try:
				if os.path.isdir( directory ):
					os.path.walk(directory, add, ())
			except (IOError, OSError, TypeError, ValueError):
				pass
		return set

def findFonts(paths = None):
	"""Find fonts in paths, or the system paths if not given

	XXX Doesn't current support OS-X system paths
	"""
	files = {}
	if paths is None:
		if sys.platform == 'win32':
			fontDirectory = win32FontDirectory()
			paths = [
				fontDirectory,
			]
			# now get all installed fonts directly...
			for f in win32InstalledFonts(fontDirectory):
				# yes, it's inefficient, the interface
				# for win32InstalledFonts really should be
				# using sets, not lists
				files[f] = 1
		else:
			paths = linuxFontDirectories()
	elif isinstance( paths, (str, unicode)):
		paths = [paths]
	for path in paths:
		for file in glob.glob( os.path.join(path, '*.ttf')):
			files[os.path.abspath(file)] = 1
	return files.keys()

if __name__ == "__main__":
	print 'linux font directories', linuxFontDirectories()
	print 'font names', findFonts()




"""Heuristic attempting to extract descriptions from font names

This should only be used as a backup in case there is
no proper available querying mechanism (i.e. fonttools),
and even then, we really would rather avoid using this
extremely fragile mechanism.

Basically what happens is that a name compression function
is run over the font name, as seen in the Windows Registry,
producing a base name and a set of modifiers, with the
modifiers being the weight (as a string) and a flag
indicating whether the font appears to be italicised.

XXX This is English-specific and entirely dependent on the
	naming conventions commonly seen when naming fonts, so
	there will be a considerable number of cases where the
	resulting name and flags will be incorrect.
"""
WEIGHTS = [
	'demibold', 'extrabold', 'semibold', 'ultrabold', 'bold',
	'ultralight', 'extralight', 'light',
	'heavy',
	'medium',
	'thin',
]
IGNORE_MODIFIERS = [
	"unicode",
	"regular",
	"normal",
	"bt",
	"let",
##	"mt",
	'itc',
	"(truetype)",
	'ft',
	'win95bt',
	'plain:1.0',
	'plain',
	'plain:',
]
WEIGHT_SYNONYMS = [
	('lt', 'light'),
	('lite', 'light'),
	('bolditalic','bold'),
	('demi', 'demibold'),
	('semi', 'semibold'),
]
WEIGHT_MODIFIERS = [
	'extra', 'ultra'
]
	

ITALIC_INDICATORS = [
	'italic', 'ital','itali', 'it', 'bolditalic',
]

def interpretModifiers( name ):
	"""Heuristic attempt to get weight and italic data from font-name

	return base_font_name, (weight, italic)
	"""
	name = name.replace('_',' ').replace('-',' ').lower().split()
	weight = "normal"
	for w in WEIGHTS:
		if w in name:
			weight = w
			while w in name:
				name.remove( w )
			break
	for w,syn in WEIGHT_SYNONYMS:
		if w in name:
			weight = syn
			while w in name:
				name.remove( w )
	if weight in ('bold','light'):
		# "extra bold", "extra light", etceteras
		for w in WEIGHT_MODIFIERS:
			if w in name:
				weight = w+weight
				while w in name:
					name.remove( w )
				break
	# XXX should look for "extra" and "ultra" w/out bold/light,
	# but that that point we're basically just shooting blind
	# as to which it is...
	for i in ITALIC_INDICATORS:
		if i in name:
			italic = 1
			while i in name:
				name.remove( i )
			break
	else:
		italic = 0
	for i in IGNORE_MODIFIERS:
		while i in name:
			name.remove( i )
	name = " ".join( name )
	if not name.strip():
		# MT Extra
		raise ValueError( "name reduced to null by name compression" )
		return None,(None,None)
	return name, (weight, italic)


def add( name, file ):
	"""Add a font with name and file to this module's registry"""
	name, style = interpretModifiers( name )
	if name is not None:
		SYSTEM_FONTS.setdefault(name, {})[style] = file
	return name, style
def get( name, style=None ):
	"""Get a font by name and optional style

	Style defaults to ("normal", 0)
	"""
	name, tstyle = interpretModifiers( name )
	if tstyle != style and style is None:
		style = tstyle
	set = SYSTEM_FONTS.get( name )
	if set:
		if set.has_key( style ):
			return set[style]
		# try match on italics
		for key, value in set.items():
			if key[1] == style[1]:
				return value
		# try match on weight
		for key, value in set.items():
			if key[0] == style[0]:
				return value
		# well, there's at least one font here, use it...
		for key, value in set.items():
			return value
	return None



#!/usr/bin/env python
"""Demo script to print ordered set of system fonts"""
from ttfquery import describe, findsystem
import sys, traceback

def buildTable( filenames=None, failureCallback=None ):
	"""Build table mapping {family:(font:{modifiers:(name,file)})}

	filenames -- if provided, list of filenames to scan,
		otherwise the full set of system fonts provided
		by findsystem will be used.
	failureCallback -- if provided, a function taking three
		arguments, the failing filename, an error-type code,
		and the error object.  If processing should stop,
		raise an error.
		codes:
			0 -- couldn't open the font file
			1 -- couldn't find modifiers in the font file
			2 -- couldn't find font-name in the font file
			3 -- couldn't find the generic family specifier
				for the font
	"""
	if filenames is None:
		filenames = findsystem.findFonts()
	table = {}
	for filename in filenames:
		try:
			font = describe.openFont(filename)
		except Exception, err:
			if failureCallback:
				failureCallback( filename, 0, err )
		else:
			try:
				modifiers = describe.modifiers( font )
			except (KeyError,AttributeError), err:
				if failureCallback:
					failureCallback( filename, 1, err )
				modifiers = (None,None)
			try:
				specificName, fontName = describe.shortName( font )
			except (KeyError,AttributeError), err:
				if failureCallback:
					failureCallback( filename, 2, err )
			else:
				try:
					specifier = describe.family(font)
				except KeyError:
					if failureCallback:
						failureCallback( filename, 3, err )
				else:
					table.setdefault(
						specifier,
						{}
					).setdefault(
						fontName,
						{}
					)[modifiers] = (specificName,filename)
	return table

def interactiveCallback( file, code, err ):
	"""Simple error callback for interactive use"""
	sys.stderr.write(
		'ERR on file %r (code %s):\n'%(
			file, code,
		)
	)
	traceback.print_exc( file=sys.stderr )

		

if __name__ == "__main__":
	import time
	t = time.clock()
	if sys.argv[1:]:
		directories = sys.argv[1:]
		files = findsystem.findFonts(directories)
	else:
		files = None
	table = buildTable(files, failureCallback=interactiveCallback)
	t = time.clock()-t
	keys = table.keys()
	keys.sort()
	for fam in keys:
		print '_________________________'
		print fam
		fnts = table[fam].items()
		fnts.sort()
		for fnt,modset in fnts:
			mods = modset.keys()
			mods.sort()
			mods = ",".join([ '%s%s'%( w, ['','(I)'][i&1]) for (w,i) in mods])
			print '    ',fnt.ljust(32), '--', mods
	print 'Scan took %s seconds CPU time'%(t,)



"""Glyph-specific queries on font-files"""
from ttfquery import describe
try:
	from OpenGLContext.debug.logs import text_log
except ImportError:
	text_log = None

def glyphName( font, char, encoding=None, warnOnFailure=1 ):
	"""Retrieve the glyph name for the given character

	XXX
		Not sure what the effect of the Unicode mapping
		will be given the use of ord...
	"""
	cmap = font['cmap']
	if encoding is None:
		encoding = describe.guessEncoding( font )
	table = cmap.getcmap( *encoding )
	glyfName = table.cmap.get( ord(char))
	if glyfName is None:
		glyfName = table.cmap.get( -1)
		if glyfName is None:
			glyfName = font['glyf'].glyphOrder[0]
			if text_log and warnOnFailure:
				text_log.warn(
					"""Unable to find glyph name for %r, in %r using first glyph in table (%r)""",
					char,
					describe.shortName(font),
					glyfName
				)
	return glyfName

def width( font, glyphName ):
	"""Retrieve the width of the giving character for given font

	The horizontal metrics table provides both the
	width and the left side bearing, we should really
	be using the left side bearing to adjust the
	character, but that's a later project.
	"""
	try:
		return font['hmtx'].metrics[ glyphName ][0]
	except KeyError:
		raise ValueError( """Couldn't find glyph for glyphName %r"""%(
			glyphName,
		))

def lineHeight( font ):
	"""Get the base-line to base-line height for the font

	XXX
		There is some fudging going on here as I
		workaround what appears to be a problem with the
		specification for sTypoDescender, which states
		that it should normally be a negative value, but
		winds up being positive in at least one font that
		defines points below the zero axis.

	XXX The entire OS/2 table doesn't appear in a few
		fonts (symbol fonts in particular), such as Corel's
		BeeHive and BlackLight 686.
	"""
	return charHeight(font) + font['OS/2'].sTypoLineGap

def charHeight( font ):
	"""Determine the general character height for the font (for scaling)"""
	ascent = font['OS/2'].sTypoAscender
	descent = font['OS/2'].sTypoDescender
	if descent > 0:
		descent = - descent
	return ascent - descent




#!/usr/bin/env python
"""Query for font-members of a particular family"""

if __name__ == "__main__":
	usage ="""ttffamily MAJOR [MINOR]

	Will create a registry file font.cache if it doesn't
	already exist, otherwise will just use the existing
	cache.  See ttffiles.py for updating the cache.
	"""
	from ttfquery._scriptregistry import registry
	import sys, os
	if sys.argv[1:2]:
		major = sys.argv[1]
		if sys.argv[2:3]:
			minor = sys.argv[2]
		else:
			minor = None
	for fontName in registry.familyMembers( major, minor ):
		print 'F', fontName
		for specific in registry.fontMembers( fontName ):
			print ' ', registry.fontFile( specific )
	
	



from ttfquery import ttffiles
import os, sys

### more robust registry-file location by John Hunter...
if os.environ.has_key('HOME'):
    registryFile = os.path.join( os.environ['HOME'], ".font.cache")
else:
	# OpenGLContext uses the Application Data directory for win32,
	# should consider porting that code here...
    registryFile = os.path.join( os.path.split(__file__)[0], "font.cache")

# make sure we can write to the registryFile
if not os.path.exists(registryFile):
    try: fh = file(registryFile, 'w')
    except IOError:
        print >>sys.stderr, 'Could not open registry file %r for writing' % registryFile
        raise
    else:
        fh.close()
        os.remove(registryFile)

def _getRegistry():
	if os.path.isfile( registryFile ):
		registry = ttffiles.load( registryFile )
	else:
		registry = ttffiles.Registry()
		sys.stderr.write( """Scanning for system fonts...\n""" )
		new,failed = registry.scan( printErrors = 1, force = 0)
		sys.stderr.write( """Scan complete. Saving to %r\n"""%(registryFile,) )
		registry.save(registryFile)
	return registry
registry = _getRegistry()



#!/usr/bin/env python
"""Installs ttfquery using distutils

Run:
	python setup.py install
to install the package from the source archive.
"""

if __name__ == "__main__":
	import sys,os, string
	from distutils.sysconfig import *
	from distutils.core import setup

	##############
	## Following is from Pete Shinners,
	## apparently it will work around the reported bug on
	## some unix machines where the data files are copied
	## to weird locations if the user's configuration options
	## were entered during the wrong phase of the moon :) .
	from distutils.command.install_data import install_data
	class smart_install_data(install_data):
		def run(self):
			#need to change self.install_dir to the library dir
			install_cmd = self.get_finalized_command('install')
			self.install_dir = getattr(install_cmd, 'install_lib')
			# should create the directory if it doesn't exist!!!
			return install_data.run(self)
	##############

	### Now the actual set up call
	setup (
		name = "TTFQuery",
		version = "0.2.6",
		description = "FontTools-based package for querying system fonts",
		author = "Mike C. Fletcher",
		author_email = "mcfletch@users.sourceforge.net",
		url = "http://members.rogers.com/mcfletch/",
		license = "BSD-style, see license.txt for details",

		package_dir = {
			'ttfquery':'.',
		},
		packages = [
			'ttfquery', 
		],
	)
	
